#!/usr/bin/gawk -f
##################################################################
# The CUDA BE generator
#
#Copyright 2008 Pjotr Kourzanov
#
#This file is part of LIME.
#
#LIME is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License version 2
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#LIME is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with LIME.  If not, see <http://www.gnu.org/licenses/>.
##################################################################
BEGIN { if (!OP) OP="dump"
       	if (!STDERR) STDERR="/dev/stderr"
	state=""  #top
}

func handle(tag,body) { return handle_gxf(tag,handle_types(tag,body)) }

END {
    merge_types()

    if (OP=="run.cu") for (f in files) if (f ~ /.gxf$/) genrun_cu(f)
    if (OP=="edges.c") for (f in files) if (f ~ /.gxf$/) genedges_c(f)
    if (OP=="kernel.h") for (f in files) if (f ~ /.gxf$/) genkernel_h(f)
    if (OP=="kernel.c") for (f in files) if (f ~ /.gxf$/) genkernel_c(f)
}

func genrun_cu(f, i,pr,fid,n,c,j,sn,pv,k,fs,fn,fp,tn,tp,z,BLOCKS,THREADS,blocks,threads) {
    pr=gensub(/\.[a-z]+$/,"","g",f)
    print "#include <cuda.h>"
    print "#include <stdio.h>"
    print "#include <cuda-debug.h>"
    BLOCKS=1
    THREADS=1
    print "#define BLOCKS",BLOCKS
    print "#define THREADS",THREADS
    for (i=1; i<=nnodes[f]; i++) {
    	n=nodes[f,i]
	# don't include header for unconnected nodes
	if (!fanout[f,n] && !fanin[f,n]) continue
	print "#include","\"" n "-kernel.h\""
    }
    print "int run(int argc, char* argv[]) {"
    # task state variables
    for (i=1; i<=nstatefuls[f]; i++) {
	n=statefuls[f,i]
	printf space(2) comp_state(n) " (*" n "_state)[%i][%i];\n",BLOCKS,THREADS
    }
    # buffers
    print space(2) "// device buffers"
    for (i=1; i<=nedges[f]; i++) {
	split(edge_from[f,i],fr,SUBSEP); fn=fr[1]; fp=fr[2]
	split(edge_to[f,i],to,SUBSEP); tn=to[1]; tp=to[2]
	if (is_state(f,fn,"",fp)) continue
	printf space(2) port_type[f,fn,fp] " (*" comp_buf(fn,fp,tn,tp) ")[" port_size[f,fn,fp] "*%i*%i];\n",BLOCKS,THREADS
	printf space(2) "cuda(Malloc((void**)&%s,sizeof(*%s)));\n",
	      comp_buf(fn,fp,tn,tp),comp_buf(fn,fp,tn,tp)
    }
    # state initialization, for constructors that are linked to the
    # main
    print space(2) "// state initialization"
    for (i=1; i<=nstatefuls[f]; i++) {
	sn=statefuls[f,i]
	printf space(2) "cuda(Malloc((void**)&%s_state,sizeof(*%s_state)));\n", sn,sn
	for (j=1; j<=nr_states(f,sn); j++) 
	{
	    p=stateports[f,sn,"read",j]
	    c=init_node[f,sn,p]
	    for (k=1; k<=nnodes[c]; k++) {
		n=nodes[c,k]
	    	args=""
	    	for (z=1; z<=nports[c,n]; z++) {
	  	  pv=ports[c,n,z]
	  	  args=append(args,
	      		(port_const[c,n,pv] ? "(const " port_type[c,n,pv] "*)" : "")\
	      		dump_portvar(f,sn,n,pv,"(*" sn "_",")[i][j]."), ", ")
	    	}
		printf space(2) "for (int i=0; i<%i; i++) for (int j=0; j<%i; j++) {\n",BLOCKS,THREADS
	    	print space(4) scpe(c,n) "(" args ");"
	    	for (z=1; z<=noutports[c,n]; z++) {
	    	  pv=outports[c,n,z]
	     	  printf space(4) "%s_state=%s_BUSY;\n",
			 dump_portvar(f,sn,n,pv,"(*" sn "_",")[i][j]."),sn
	    	}
		print space(2) "}"
	    }
	}
    }
    # task creation
    print space(2) "// task creation"
    for (i=1; i<=nnodes[f]; i++) {
	n=nodes[f,i]
	# don't create a task for unconnected nodes
	if (!is_connected(f,n)) continue
	if (is_structor(f,n)) continue
	args=""
	for (k=1; k<=nports[f,n]; k++) {
	    p=ports[f,n,k]
	    if (is_state(f,n,n,p))
	  	args=append(args,
	      		"*" n "_state", ", ")
	    else
	  	args=append(args,
	      		(port_const[f,n,p] ? "(const " port_type[f,n,p] "*)" : "")\
	      		"*" comp_buf(n,p,peer_node[f,n,p],peer_port[f,n,p]), ", ")

	}
	blocks[n]=BLOCKS
	threads[n]=THREADS
	printf space(2) "%s <<<dim3(%i),dim3(%i)>>>(%s);\n",comp_kernel(n),BLOCKS,THREADS,args
    }
    # state deinitialization
    print space(2) "// state deinitialization"
    for (i=1; i<=nstatefuls[f]; i++) {
	sn=statefuls[f,i]
	for (j=1; j<=nr_dstrstates(f,sn); j++) 
	{
	    p=stateports[f,sn,"write",j]
	    c=deinit_node[f,sn,p]
	    for (k=1; k<=nnodes[c]; k++) {
		n=nodes[c,k]
	    	args=""
	    	for (z=1; z<=nports[c,n]; z++) {
	  	  pv=ports[c,n,z]
	  	  args=append(args,
	      		(port_const[c,n,pv] ? "(const " port_type[c,n,pv] "*)" : "")\
	      		dump_portvar(f,sn,n,pv,"(*" sn "_",")[i][j]."), ", ")
	    	}
		printf space(2) "for (int i=0; i<%i; i++) for (int j=0; j<%i; j++) {\n",BLOCKS,THREADS
	    	print space(4) scpe(c,n) "(" args ");"
	    	for (z=1; z<=noutports[c,n]; z++) {
	    	  pv=outports[c,n,z]
	     	  printf space(4) "%s_state=%s_IDLE;\n",
			 dump_portvar(f,sn,n,pv,"(*" sn "_",")[i][j]."),sn
	    	}
		print space(2) "}"
	    }
	}
    }
    print space(2) "// release buffer memory"
    for (i=1; i<=nedges[f]; i++) {
	split(edge_from[f,i],fr,SUBSEP); fn=fr[1]; fp=fr[2]
	split(edge_to[f,i],to,SUBSEP); tn=to[1]; tp=to[2]
	if (is_state(f,fn,"",fp)) continue
	printf space(2) "cuda(Free(%s));\n",
	      comp_buf(fn,fp,tn,tp)
    }
    print space(2) "// release state memory"
    for (i=1; i<=nstatefuls[f]; i++) {
	sn=statefuls[f,i]
	printf space(2) "cuda(Free(%s_state));\n", sn
    }
    print space(2) "return EXIT_SUCCESS;"
    print "}"
    for (i=1; i<=nnodes[f]; i++) {
    	n=nodes[f,i]
	# don't include header for unconnected nodes
	if (!fanout[f,n] && !fanin[f,n]) continue
	print "#define BLOCKS",blocks[n]
	print "#define THREADS",threads[n]
	print "#include","\"" n "-kernel.c\""
    }
}

func genkernel_h(f, i,k,args,n,fs,p,defined) {
    fs=PE "_h_defined" 
    print "#ifndef",fs
    print "#define",fs
    delete included
    while (getline < IMPORTS) {
	#gsub(/.*\//,"")
	if (included[$0]) continue
    	print "#include \"" $0 ".h\""
	included[$0]=1
    }
    close(IMPORTS)
    n=PE
    if ((f,n) in node_idx); else fail("node " n " not found in " f)
    # {con,de}structors do not require a shell header, besides the
    # forward declaration for the actual call
    if (is_structor(f,n)) {
	dump_prototypes(f,n)
    	print "#endif//" fs
	return
    }

    if (nr_states(f,n)) {
      for (i=1; i<=nr_states(f,n); i++) {
    	p=stateports[f,n,"read",i]
        print "enum","_" n "_" p "_state {\n\t",
        	n "_IDLE = 0,\n\t",
      		n "_BUSY,\n\t",
      		n "_STOP\n",
    		"};"
      }
      print comp_state(n),"{"
      # only generate state instance for incoming ports
      for (i=1; i<=nr_states(f,n); i++) {
    	p=stateports[f,n,"read",i]
	if (p in defined) continue
    	print space(2) "enum","_" n "_" p "_state ", p "_state:2;" 
    	print space(2) port_type[f,n,p],p "[" port_size[f,n,p] "];"
	defined[p]=1
      }
      print "};"
    }
    for (i=1; i<=nnodes[f]; i++) if ((c=nodes[f,i])==PE) {
    	# {con,de}structors do not require a shell
	if (is_structor(f,c)) continue

	args=""
	for (k=1; k<=nports[f,c]; k++) {
	    p=ports[f,c,k]
	    if (is_state(f,c,c,p))
	  	args=append(args,
	      		(port_const[f,c,p] ? "const " : "") comp_state(c)\
			" " p "[BLOCKS][THREADS]", ", ")
	    else
	  	args=append(args,
	      		(port_const[f,c,p] ? "const " : "") port_type[f,c,p]\
			" " p "[" port_size[f,c,p] "*BLOCKS*THREADS]", ", ")

	}

    	printf "__global__ void %s(%s);\n",comp_kernel(c),args
    }
    print "#endif//" fs
}

func genkernel_c(f, i,c,n,k) {
    print "#include \"" PE "-kernel.h\""
    print "#include \"" PE ".mod.c\""
    # for all actors
    for (i=1; i<=nnodes[f]; i++) if ((c=nodes[f,i])==PE) {
    	# {con,de}structors do not require a shell
	if (is_structor(f,c)) continue

	args=""
	for (k=1; k<=nports[f,c]; k++) {
	    p=ports[f,c,k]
	    if (is_state(f,c,c,p))
	  	args=append(args,
	      		(port_const[f,c,p] ? "const " : "") comp_state(c)\
			" " p "[BLOCKS][THREADS]", ", ")
	    else
	  	args=append(args,
	      		(port_const[f,c,p] ? "const " : "") port_type[f,c,p]\
			" " p "[" port_size[f,c,p] "*BLOCKS*THREADS]", ", ")

	}

    	printf "__global__ void %s(%s) {\n",comp_kernel(c),args
	# for all sub-actors
	for (k=1; k<=nnodes[c]; k++)
	{
		n=nodes[c,k]
		do_call(f,c,n)
	}
	print "}"
    }
}
func do_call(f,c,n, p,args,cond,chng,j) {
  for (j=1; j<=nports[c,n]; j++) {
    p=ports[c,n,j]
    pv=dump_portvar(f,c,n,p)
    if ((c,n,p) in cons_idx) {
	    cond=append(cond,pv "_state==" c "_IDLE","&&")
	    chng=append(chng,pv "_state=" c "_BUSY;","\n")
    }
    if ((c,n,p) in dstr_idx) {
	    cond=append(cond,pv "_state==" c "_STOP","&&")
	    chng=append(chng,pv "_state=" c "_IDLE;","\n")
    }
    if (is_state(f,c,n,p))
    	args=append(args,
		(port_const[c,n,p] ? "(const " port_type[c,n,p] "*)" : "")\
		p "[blockIdx.x][threadIdx.x]." fold_portvar(f,c,n,p), ", ")
    else
    	args=append(args,
		(port_const[c,n,p] ? "(const " port_type[c,n,p] "*)" : "")\
		"&" pv "[" port_size[c,n,p] "*(blockIdx.x*blockDim.x+threadIdx.x)]", ", ")
  }
  if (cond && chng) {
	print space(2) "if (" cond ") {"
	print space(4) scpe(c,n) "(" args ");"
	print space(4) chng
	print space(2) "}"
  } else  print space(2) scpe(c,n) "(" args ");"
}
func comp_buf(fn,fp,tn,tp) {
	if (port_const[f,tn,tp])
		return fn "_" fp "__" tn "_" tp 
	return tn "_" tp "__" fn "_" fp 
}
func comp_kernel(n,pr,sep) { return (pr ? pr (sep ? sep : "_") : "") n "_kernel" }

@include gen-shell.awk
@include gxf-parser.awk
@include types-parser.awk
