/****************************************************************
Simple sink to demonstrate LIME-CUDA.
State is constructed externally.

Copyright 2008 Pjotr Kourzanov

This file is part of LIME.

LIME is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

LIME is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with LIME.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/
#include CUDA
#include CUDA_stdio

#if defined(__DEVICE_EMULATION__)
#include <assert.h>
#endif

struct source_state;

static int CheckCount(unsigned int size, const unsigned int *buf, unsigned int *pCount);

actor process(const unsigned int buf[restrict BUF_SIZE],
	      const struct source_state inst[1],
	      struct source_state outs[1])
{

	int s=CheckCount(BUF_SIZE,buf,&outs->count);

#if defined(__DEVICE_EMULATION__)
	if (s) printf("check OK\n");
	else printf("check FAILED\n");
	assert(s);
#endif
}

helper static int CheckCount(unsigned int size, 
		const unsigned int *buf,
		unsigned int *pCount)
{
	unsigned int i;
	for (i=0; i < size; i++)
	{
		if (buf[i] != (*pCount))
		{
			return 0;
		}
		if (buf[i] >> 23)
		{
			return 0;
		}
		/* Wrap on signed 24 bits value */
		atomicInc(pCount,UINT_MAX);
		atomicAnd(pCount,0x7fffff);		
	}
	return 1;
}		
