##################################################################
# This AWK script implements generation of description package 
# given a set of XML files that describe the architecture platform,
# and the application use-case
#
#Copyright 2008 Pjotr Kourzanov
#
#This file is part of LIME.
#
#LIME is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License version 2
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#LIME is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with LIME.  If not, see <http://www.gnu.org/licenses/>.
##################################################################
BEGIN {
	if (!rname) rname="radio"
	if (!rnum) rnum="0x0BABA000U"

	print "#include <Radio.h>"
	printf "#include <%s-comps.h>\n",rname
	printf "#include <%s-dspcfg.h>\n",rname
	printf "radio_t %s = { setf(radio_id,%s),\n",rname,rnum
}
END {
	for (f in files) if (f ~ /.gxf$/) {
		gxf=f
		break
	}
	for (i=1; i<=ncomp; i++)
	{
		if (!cclass[comp[i]]) continue
		nmcomp++
	}
	printf space(2) "setf(nr_of_comps,%u),\n",nmcomp
	print space(2) "setf(comp_list,scpe("
	j=0
	for (i=1; i<=ncomp; i++)
	{
		if (!cclass[comp[i]]) continue
	    	print space(4) "/*",comp[i],"*/"
		printf space(4) "seta(%i, scpe(%s,\n"\
			     "                 setf(cores, scpe(setf(%s,\n"\
			     "                                  scpe(setf(type, %s_TYPE),\n"\
			     "                                       setf(  id, %s))))))),\n",
			j++,put_comp_id(i),
			cclass[comp[i]],
			ctype[comp[i]],
			cid[comp[i]]
	}
	print space(2) ")),"
	print space(2) "setf(comp_req_list,scpe("
	z=0
	for (i=1; i<=ncomp; i++) 
	{
	    if (!cclass[comp[i]]) continue
	    print space(4) "/*",comp[i],"*/"
	    printf space(4) "seta(%i, scpe(setf(nr_of_reqs,%u),\n"\
  	            "                 setf(req_list, %s\n",
		       z++,nreq[comp[i]],(nreq[comp[i]] ? "scpe(" : "{}")
	    for (j=1; j<=nreq[comp[i]]; j++) {
		printf space(20) "seta(%i, scpe(setf(type,%s_REQ),\n"\
		       space(20) "             setf(param_list,scpe(\n",
		      		j-1,
				gensub(/.*:/,"","g",req_resource[comp[i],j])
		n=split(unquote(req_quantity[comp[i],j]),r)
		if (req_address[comp[i],j]) r[++n]=req_address[comp[i],j]
		for (k=1; k<=n; k++)
		   if (comptype[i]=="fifo") {
		   	split(edge_from[gxf,id_edge[comp[i]]],fr,SUBSEP)
			printf space(36) "seta(%u, %s*unitsize(%s)),\n",
				k-1,
				unquote(r[k]),
				port_type[gxf,fr[1],fr[2]]
		   } else
			printf space(36) "seta(%u, %s),\n",
				k-1,
				unquote(r[k])
		print space(33) ")))),"
	    }
	    printf space(4) "%s))),\n",(nreq[comp[i]] ? ")" : "")
	}
	print space(2) ")),"
 	genedges(gxf)
	print "};"
}
func handle_core(t,b) {
    if (t == "@id") return id[lev]=b
    
    if (relax) {
#    	if (DEBUG) print "CORE: ignoring",t,"in",state >STDERR
    	return b
    }

    fail("CORE: unknown tag " t)
}
func handle_arch(t,b) {

    if (handle_xlink(t,b)) return b
    
    if (t == "@id") return id[lev]=b

    if (t ~ /^(type|rtype)$/ && type[lev+1]=="locator") {
	ltype[++nltype]=label[lev+1]
	lname[label[lev+1]]=href[2]
	delete type[lev+1]; delete label[lev+1]; delete href
	return b
    }

    if (t ~ /^(core|cores|memory|resource)$/ && type[lev+1]=="resource") {
	rid[label[lev+1]]=id[lev+1]
	delete type[lev+1]; delete label[lev+1]; delete id[lev+1]
	return b
    }

    if (t ~ /^(type-is|rtype-is)$/ && type[lev+1]=="arc") {
	typeof[from[lev+1]]=lname[to[lev+1]]
	classof[from[lev+1]]=to[lev+1]
#delete lname
	delete type[lev+1]; delete from[lev+1]; delete to[lev+1]
        return b
    }

    if (relax) {
#    	if (DEBUG) print "ARCH: ignoring",t,"in",state >STDERR
    	return b
    }

    fail("ARCH: unknown tag " t)
}
func handle_maps(t,b, a) {

    if (handle_xlink(t,b)) return b

    if (t == "resource" && type[lev+1]!="locator") return resource[lev]=b
    if (t == "quantity") return quantity[lev]=b
    if (t == "address") return address[lev]=b
    
    if (t ~ /^(core|cores|memory|actor|fifo|resource)$/ && type[lev+1]=="locator") {
	if (t=="actor" || t=="fifo") {
		comp[++ncomp]=href[2]
		comptype[ncomp]=t
	}
	if (t=="resource") rid[label[lev+1]]=href[2]
	if (t=="cores") {
		rid[label[lev+1]]=href[2] "_CORES"
		typeof[label[lev+1]]=href[2]
		classof[label[lev+1]]=label[lev+1]
	}
	delete type[lev+1]; delete label[lev+1]; delete href
	return b
    }

    if (t == "core-of-node" && type[lev+1]=="arc") {
        if (typeof[to[lev+1]]!="MEM") {	
	    cclass[from[lev+1]]=classof[to[lev+1]]
	    ctype[from[lev+1]]=typeof[to[lev+1]]
	    cid[from[lev+1]]=rid[to[lev+1]]
#delete typeof; delete rid; delete classof
	}
	delete type[lev+1]; delete from[lev+1]; delete to[lev+1]
	return b
    }
    
    if (t == "provides" && type[lev+1]=="arc") {
	delete resource[lev+1]; delete quantity[lev+1]; delete address[lev+1]
	delete type[lev+1]; delete from[lev+1]; delete to[lev+1]
	return b
    }
    
    if (t == "requires" && type[lev+1]=="arc") {
	req_resource[from[lev+1],++nreq[from[lev+1]]]=rid[to[lev+1]] #resource[lev+1]
	req_quantity[from[lev+1],nreq[from[lev+1]]]=quantity[lev+1]
	if (address[lev+1]) req_address[from[lev+1],nreq[from[lev+1]]]=address[lev+1]
	delete resource[lev+1]; delete quantity[lev+1]; delete address[lev+1]
	delete type[lev+1]; delete from[lev+1]; delete to[lev+1]
	return b
    }
    
    if (relax) {
#    	if (DEBUG) print "MAPS: ignoring",t,"in",state >STDERR
    	return b
    }

    fail("MAPS: unknown tag " t)
}

func handle(t,b) { return handle_maps(t,handle_gxf(t,handle_arch(t,handle_core(t,b)))) }

@include gxf-parser.awk
@include gen-shell.awk
